﻿using System.Collections.Generic;
using System.Linq;

using Buddy.Wildstar.Engine;
using Buddy.Wildstar.Game;
using Buddy.Wildstar.Game.Actors;

using log4net;

namespace GoodStalker
{
	public class Targeting
	{
		private static ILog Log = LogManager.GetLogger("Default Routine [Targeting]");

		// Technical note: PerFrameCachedValue is automatically updated every "Frame" (or Tick) the bot uses. This allows us to process something once, and only once per frame, saving massively
		// on computation time.

		private static readonly PerFrameCachedValue<Actor> BestTargetCache = new PerFrameCachedValue<Actor>(() => SortedCombatTargets.FirstOrDefault());
		private static readonly PerFrameCachedValue<IEnumerable<Actor>> SortedTargetsCache = new PerFrameCachedValue<IEnumerable<Actor>>(() => CombatTargets.OrderByDescending(Weigh));
		private static readonly PerFrameCachedValue<List<Actor>> CombatTargetsCache = new PerFrameCachedValue<List<Actor>>(() =>
		{
			// Ok, so, this is so we can handle including mobs in combat with our pets.
			// This is a *major* fix for engineers.
			HashSet<uint> friendlyGuids = new HashSet<uint>();
			friendlyGuids.Add(GameManager.LocalPlayer.Guid);
			if (GameManager.CurrentPets.Count != 0)
				friendlyGuids.UnionWith(GameManager.CurrentPets.Select(g => g.Pet.Guid));

			return GameManager.Actors.Values.Where(a =>
			{
				// Ignore null actors (should never happen, but sanity check)
				if (a == null)
				{
					return false;
				}
				// Make sure the actor is backed by a valid memory address
				if (!a.IsValid)
				{
					return false;
				}

				// Obviously can't attack dead stuff!
				if (a.IsDead)
				{
					return false;
				}

				var ci = a.CreatureInfo;
				if (ci != null)
				{
					// TEMPORARY: Possibly "CanBeAttacked" ?
					if ((ci.Flags & 1) != 0)
						return false;
				}

				// Friendly = 2, Unknown = 3+
				// Hostile = 0, Neutral = 1
				// We only want to "attack" hostile or neutral actors
				if (a.Disposition >= Disposition.Friendly)
				{
					return false;
				}

				// If it's targeting us, we want to include it. Even if the combat checks below fail.
				if (a.CurrentTargetGuid != 0 && friendlyGuids.Contains(a.CurrentTargetGuid))
				{
					return true;
				}

				// The thing isn't in combat, so ignore it.
				if (!a.IsInCombat)
				{
					return false;
				}

				// If it's tagged by us, or our group, go ahead and include it. Even if it's not our main priority.
				if (a.TaggedByLocalPlayer || a.TaggedByLocalPlayerGroup)
				{
					return true;
				}

				return false;
			}).ToList();
		});

		private static Actor _pullTarget;

		/// <summary>
		/// The current target passed into the Pull call of the routine.
		/// This is used for forcing combat with certain things (such as herbs, which are required by game mechanics)
		/// </summary>
		public static Actor PullTarget
		{
			get
			{
				if (!_pullTarget.IsValid)
					_pullTarget = null;
				return _pullTarget;
			}
			set { _pullTarget = value; }
		}

		/// <summary>
		/// The current best kill target.
		/// </summary>
		public static Actor BestTarget
		{
			get
			{
				if (PullTarget != null && PullTarget.IsValid && !PullTarget.IsDead)
					return PullTarget;
				return BestTargetCache;
			}
		}

		/// <summary>
		/// A cached list of all available combat targets, without any weight-based sorting applied.
		/// </summary>
		public static List<Actor> CombatTargets { get { return CombatTargetsCache; } }

		/// <summary>
		/// A weight-sorted list of combat targets.
		/// </summary>
		public static IEnumerable<Actor> SortedCombatTargets { get { return SortedTargetsCache.Value; } }

		private static float Weigh(Actor actor)
		{
			float w = 0f;

			if (actor == null || !actor.IsValid)
				return float.MinValue;

			// Lower health actors get higher weight.
			// Subtract their health*2 from 0, and the lower their health, the higher their weight.
			w -= actor.HealthPercent * 2f;
			// Shields are effectively regenerative health. For monsters, it doesn't seem to regen (excluding boss fights)
			// So it can be treated as an extra health buffer.
			w -= actor.ShieldPercent * 2f;
			// Same as the health & shields
			w -= actor.AbsorptionPercent * 1.5f;

			// The more interrupt armor it has, the more we probably want to kill it.
			// TODO: This may not be a good idea. Might be useful to do an extra check elsewhere for interrupt armor when we need to use interrupts
			w += actor.InterruptArmor * 2;

			// The further away it is, the lower the weight. Similar to HP.
			w -= actor.Distance * 1.5f;

			// Extra weight if it's the current pull target.
			if (PullTarget != null && PullTarget.IsValid && PullTarget.Guid == actor.Guid)
				w += 200;

			// Try to keep our current target, targeted.
			if (GameManager.LocalPlayer.CurrentTargetGuid == actor.Guid)
				w += 100;

			return w;
		}
	}
}